#
# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (c) 2013-2025, John McNamara, jmcnamara@cpan.org
#
__version__ = "3.2.8"
__VERSION__ = __version__
from .workbook import Workbook  # noqa

__all__ = ["Workbook"]



###############################################################################
#
# Format - A class for writing the Excel XLSX Worksheet file.
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (c) 2013-2025, John McNamara, jmcnamara@cpan.org
#

# Package imports.
from typing import Literal, Union
from warnings import warn

from xlsxwriter.color import Color

from . import xmlwriter


class Format(xmlwriter.XMLwriter):
    """
    A class for writing the Excel XLSX Format file.


    """

    ###########################################################################
    #
    # Public API.
    #
    ###########################################################################

    def __init__(self, properties=None, xf_indices=None, dxf_indices=None) -> None:
        """
        Constructor.

        """
        if properties is None:
            properties = {}

        super().__init__()

        self.xf_format_indices = xf_indices
        self.dxf_format_indices = dxf_indices
        self.xf_index = None
        self.dxf_index = None

        self.num_format = "General"
        self.num_format_index = 0
        self.font_index = 0
        self.has_font = False
        self.has_dxf_font = False

        self.bold = 0
        self.underline = 0
        self.italic = 0
        self.font_name = "Calibri"
        self.font_size = 11
        self.font_color = None
        self.font_strikeout = 0
        self.font_outline = 0
        self.font_shadow = 0
        self.font_script = 0
        self.font_family = 2
        self.font_charset = 0
        self.font_scheme = "minor"
        self.font_condense = 0
        self.font_extend = 0
        self.theme = 0
        self.hyperlink = False
        self.xf_id = 0

        self.hidden = 0
        self.locked = 1

        self.text_h_align = 0
        self.text_wrap = 0
        self.text_v_align = 0
        self.text_justlast = 0
        self.rotation = 0

        self.fg_color = None
        self.bg_color = None
        self.pattern = 0
        self.has_fill = False
        self.has_dxf_fill = False
        self.fill_index = 0
        self.fill_count = 0

        self.border_index = 0
        self.has_border = False
        self.has_dxf_border = False
        self.border_count = 0

        self.bottom = 0
        self.bottom_color = None
        self.diag_border = 0
        self.diag_color = None
        self.diag_type = 0
        self.left = 0
        self.left_color = None
        self.right = 0
        self.right_color = None
        self.top = 0
        self.top_color = None

        self.indent = 0
        self.shrink = 0
        self.merge_range = 0
        self.reading_order = 0
        self.just_distrib = 0
        self.color_indexed = 0
        self.font_only = 0

        self.quote_prefix = False
        self.checkbox = False

        # Convert properties in the constructor to method calls.
        for key, value in properties.items():
            getattr(self, "set_" + key)(value)

        self._format_key = None

    def __repr__(self) -> str:
        """
        Return a string representation of the Format instance.
        """
        return (
            f"Format("
            f"font_name={self.font_name!r}, "
            f"font_size={self.font_size}, "
            f"bold={self.bold}, "
            f"italic={self.italic}, "
            f"underline={self.underline}, "
            f"font_color={self.font_color}, "
            f"num_format={self.num_format!r}, "
            f"text_h_align={self.text_h_align}, "
            f"text_v_align={self.text_v_align}, "
            f"fg_color={self.fg_color}, "
            f"bg_color={self.bg_color}, "
            f"pattern={self.pattern}, "
            f"locked={self.locked}, "
            f"hidden={self.hidden})"
        )

    ###########################################################################
    #
    # Format properties.
    #
    ###########################################################################

    def set_font_name(self, font_name) -> None:
        """
        Set the Format font_name property such as 'Time New Roman'. The
        default Excel font is 'Calibri'.

        Args:
            font_name: String with the font name. No default.

        Returns:
            Nothing.

        """
        self.font_name = font_name

    def set_font_size(self, font_size: int = 11) -> None:
        """
        Set the Format font_size property. The default Excel font size is 11.

        Args:
            font_size: Int with font size. No default.

        Returns:
            Nothing.

        """
        self.font_size = font_size

    def set_font_color(self, font_color: Union[str, Color]) -> None:
        """
        Set the Format font_color property. The Excel default is black.

        Args:
            font_color: String with the font color. No default.

        Returns:
            Nothing.

        """
        self.font_color = Color._from_value(font_color)

    def set_bold(self, bold: bool = True) -> None:
        """
        Set the Format bold property.

        Args:
            bold: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.bold = bold

    def set_italic(self, italic: bool = True) -> None:
        """
        Set the Format italic property.

        Args:
            italic: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.italic = italic

    def set_underline(self, underline: Literal[1, 2, 33, 34] = 1) -> None:
        """
        Set the Format underline property.

        Args:
            underline: Default is 1, single underline.

        Returns:
            Nothing.

        """
        self.underline = underline

    def set_font_strikeout(self, font_strikeout: bool = True) -> None:
        """
        Set the Format font_strikeout property.

        Args:
            font_strikeout: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.font_strikeout = font_strikeout

    def set_font_script(self, font_script: Literal[1, 2] = 1) -> None:
        """
        Set the Format font_script property.

        Args:
            font_script: Default is 1, superscript.

        Returns:
            Nothing.

        """
        self.font_script = font_script

    def set_font_outline(self, font_outline: bool = True) -> None:
        """
        Set the Format font_outline property.

        Args:
            font_outline: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.font_outline = font_outline

    def set_font_shadow(self, font_shadow: bool = True) -> None:
        """
        Set the Format font_shadow property.

        Args:
            font_shadow: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.font_shadow = font_shadow

    def set_num_format(self, num_format: str) -> None:
        """
        Set the Format num_format property such as '#,##0'.

        Args:
            num_format: String representing the number format. No default.

        Returns:
            Nothing.

        """
        self.num_format = num_format

    def set_locked(self, locked: bool = True) -> None:
        """
        Set the Format locked property.

        Args:
            locked: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.locked = locked

    def set_hidden(self, hidden: bool = True) -> None:
        """
        Set the Format hidden property.

        Args:
            hidden: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.hidden = hidden

    def set_align(
        self,
        alignment: Literal[
            "left",
            "centre",
            "center",
            "right",
            "fill",
            "justify",
            "center_across",
            "centre_across",
            "distributed",
            "justify_distributed",
            "justify_distributed",
            "top",
            "vcentre",
            "vcenter",
            "bottom",
            "vjustify",
            "vdistributed",
        ],
    ) -> None:
        """
        Set the Format cell alignment.

        Args:
            alignment: String representing alignment. No default.

        Returns:
            Nothing.
        """
        alignment = alignment.lower()

        # Set horizontal alignment properties.
        if alignment == "left":
            self.set_text_h_align(1)
        if alignment == "centre":
            self.set_text_h_align(2)
        if alignment == "center":
            self.set_text_h_align(2)
        if alignment == "right":
            self.set_text_h_align(3)
        if alignment == "fill":
            self.set_text_h_align(4)
        if alignment == "justify":
            self.set_text_h_align(5)
        if alignment == "center_across":
            self.set_text_h_align(6)
        if alignment == "centre_across":
            self.set_text_h_align(6)
        if alignment == "distributed":
            self.set_text_h_align(7)
        if alignment == "justify_distributed":
            self.set_text_h_align(7)

        if alignment == "justify_distributed":
            self.just_distrib = 1

        # Set vertical alignment properties.
        if alignment == "top":
            self.set_text_v_align(1)
        if alignment == "vcentre":
            self.set_text_v_align(2)
        if alignment == "vcenter":
            self.set_text_v_align(2)
        if alignment == "bottom":
            self.set_text_v_align(3)
        if alignment == "vjustify":
            self.set_text_v_align(4)
        if alignment == "vdistributed":
            self.set_text_v_align(5)

    def set_center_across(self, align_type: None = None) -> None:
        # pylint: disable=unused-argument
        """
        Set the Format center_across property.

        Returns:
            Nothing.

        """
        self.set_text_h_align(6)

    def set_text_wrap(self, text_wrap: bool = True) -> None:
        """
        Set the Format text_wrap property.

        Args:
            text_wrap: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.text_wrap = text_wrap

    def set_rotation(self, rotation: int) -> None:
        """
        Set the Format rotation property.

        Args:
            rotation: Rotation angle. No default.

        Returns:
            Nothing.

        """
        rotation = int(rotation)

        # Map user angle to Excel angle.
        if rotation == 270:
            rotation = 255
        elif -90 <= rotation <= 90:
            if rotation < 0:
                rotation = -rotation + 90
        else:
            warn("Rotation rotation outside range: -90 <= angle <= 90")
            return

        self.rotation = rotation

    def set_indent(self, indent: int = 1) -> None:
        """
        Set the Format indent property.

        Args:
            indent: Default is 1, first indentation level.

        Returns:
            Nothing.

        """
        self.indent = indent

    def set_shrink(self, shrink: bool = True) -> None:
        """
        Set the Format shrink property.

        Args:
            shrink: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.shrink = shrink

    def set_text_justlast(self, text_justlast: bool = True) -> None:
        """
        Set the Format text_justlast property.

        Args:
            text_justlast: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.text_justlast = text_justlast

    def set_pattern(self, pattern: int = 1) -> None:
        """
        Set the Format pattern property.

        Args:
            pattern: Default is 1, solid fill.

        Returns:
            Nothing.

        """
        self.pattern = pattern

    def set_bg_color(self, bg_color: Union[str, Color]) -> None:
        """
        Set the Format bg_color property.

        Args:
            bg_color: Background color. No default.

        Returns:
            Nothing.

        """
        self.bg_color = Color._from_value(bg_color)

    def set_fg_color(self, fg_color: Union[str, Color]) -> None:
        """
        Set the Format fg_color property.

        Args:
            fg_color: Foreground color. No default.

        Returns:
            Nothing.

        """
        self.fg_color = Color._from_value(fg_color)

    # set_border(style) Set cells borders to the same style
    def set_border(self, style: int = 1) -> None:
        """
        Set the Format bottom property.

        Args:
            bottom: Default is 1, border type 1.

        Returns:
            Nothing.

        """
        self.set_bottom(style)
        self.set_top(style)
        self.set_left(style)
        self.set_right(style)

    # set_border_color(color) Set cells border to the same color
    def set_border_color(self, color: Union[str, Color]) -> None:
        """
        Set the Format bottom property.

        Args:
            color: Color string. No default.

        Returns:
            Nothing.

        """
        self.set_bottom_color(color)
        self.set_top_color(color)
        self.set_left_color(color)
        self.set_right_color(color)

    def set_bottom(self, bottom: int = 1) -> None:
        """
        Set the Format bottom property.

        Args:
            bottom: Default is 1, border type 1.

        Returns:
            Nothing.

        """
        self.bottom = bottom

    def set_bottom_color(self, bottom_color: Union[str, Color]) -> None:
        """
        Set the Format bottom_color property.

        Args:
            bottom_color: Color string. No default.

        Returns:
            Nothing.

        """
        self.bottom_color = Color._from_value(bottom_color)

    def set_diag_type(self, diag_type: Literal[1, 2, 3] = 1) -> None:
        """
        Set the Format diag_type property.

        Args:
            diag_type: Default is 1, border type 1.

        Returns:
            Nothing.

        """
        self.diag_type = diag_type

    def set_left(self, left: int = 1) -> None:
        """
        Set the Format left property.

        Args:
            left: Default is 1, border type 1.

        Returns:
            Nothing.

        """
        self.left = left

    def set_left_color(self, left_color: Union[str, Color]) -> None:
        """
        Set the Format left_color property.

        Args:
            left_color: Color string. No default.

        Returns:
            Nothing.

        """
        self.left_color = Color._from_value(left_color)

    def set_right(self, right: int = 1) -> None:
        """
        Set the Format right property.

        Args:
            right: Default is 1, border type 1.

        Returns:
            Nothing.

        """
        self.right = right

    def set_right_color(self, right_color: Union[str, Color]) -> None:
        """
        Set the Format right_color property.

        Args:
            right_color: Color string. No default.

        Returns:
            Nothing.

        """
        self.right_color = Color._from_value(right_color)

    def set_top(self, top: int = 1) -> None:
        """
        Set the Format top property.

        Args:
            top: Default is 1, border type 1.

        Returns:
            Nothing.

        """
        self.top = top

    def set_top_color(self, top_color: Union[str, Color]) -> None:
        """
        Set the Format top_color property.

        Args:
            top_color: Color string. No default.

        Returns:
            Nothing.

        """
        self.top_color = Color._from_value(top_color)

    def set_diag_color(self, diag_color: Union[str, Color]) -> None:
        """
        Set the Format diag_color property.

        Args:
            diag_color: Color string. No default.

        Returns:
            Nothing.

        """
        self.diag_color = Color._from_value(diag_color)

    def set_diag_border(self, diag_border: int = 1) -> None:
        """
        Set the Format diag_border property.

        Args:
            diag_border: Default is 1, border type 1.

        Returns:
            Nothing.

        """
        self.diag_border = diag_border

    def set_quote_prefix(self, quote_prefix: bool = True) -> None:
        """
        Set the Format quote prefix property.

        Args:
            quote_prefix: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.quote_prefix = quote_prefix

    def set_checkbox(self, checkbox: bool = True) -> None:
        """
        Set the Format property to show a checkbox in a cell.

        This format property can be used with a cell that contains a boolean
        value to display it as a checkbox. This property isn't required very
        often and it is generally easier to create a checkbox using the
        ``worksheet.insert_checkbox()`` method.

        Args:
            checkbox: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.checkbox = checkbox

    ###########################################################################
    #
    # Internal Format properties. These aren't documented since they are
    # either only used internally or else are unlikely to be set by the user.
    #
    ###########################################################################

    def set_has_font(self, has_font: bool = True) -> None:
        """
        Set the property to indicate the format has a font.

        Args:
            has_font: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.has_font = has_font

    def set_has_fill(self, has_fill: bool = True) -> None:
        """
        Set the property to indicate the format has a fill.

        Args:
            has_fill: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.has_fill = has_fill

    def set_font_index(self, font_index: int) -> None:
        """
        Set the unique font index property.

        Args:
            font_index: The unique font index.

        Returns:
            Nothing.

        """
        self.font_index = font_index

    def set_xf_index(self, xf_index: int) -> None:
        """
        Set the unique format index property.

        Args:
            xf_index: The unique Excel format index.

        Returns:
            Nothing.

        """
        self.xf_index = xf_index

    def set_dxf_index(self, dxf_index: int) -> None:
        """
        Set the unique conditional format index property.

        Args:
            dxf_index: The unique Excel conditional format index.

        Returns:
            Nothing.

        """
        self.dxf_index = dxf_index

    def set_num_format_index(self, num_format_index: int) -> None:
        """
        Set the number format_index property.

        Args:
            num_format_index: The unique number format index.

        Returns:
            Nothing.

        """
        self.num_format_index = num_format_index

    def set_text_h_align(self, text_h_align: int) -> None:
        """
        Set the horizontal text alignment property.

        Args:
            text_h_align: Horizontal text alignment.

        Returns:
            Nothing.

        """
        self.text_h_align = text_h_align

    def set_text_v_align(self, text_v_align: int) -> None:
        """
        Set the vertical text alignment property.

        Args:
            text_h_align: Vertical text alignment.

        Returns:
            Nothing.

        """
        self.text_v_align = text_v_align

    def set_reading_order(self, direction: int = 0) -> None:
        # Set the reading_order property.
        """
        Set the reading order property.

        Args:
            direction: Default is 0, left to right.

        Returns:
            Nothing.

        """
        self.reading_order = direction

    def set_valign(
        self,
        align: Literal[
            "left",
            "centre",
            "center",
            "right",
            "fill",
            "justify",
            "center_across",
            "centre_across",
            "distributed",
            "justify_distributed",
            "justify_distributed",
            "top",
            "vcentre",
            "vcenter",
            "bottom",
            "vjustify",
            "vdistributed",
        ],
    ) -> None:
        # Set vertical cell alignment. This is required by the constructor
        # properties dict to differentiate between the vertical and horizontal
        # properties.
        """
        Set vertical cell alignment property.

        This is required by the constructor properties dict to differentiate
        between the vertical and horizontal properties.

        Args:
            align: Alignment property.

        Returns:
            Nothing.

        """
        self.set_align(align)

    def set_font_family(self, font_family: int) -> None:
        """
        Set the font family property.

        Args:
            font_family: Font family number.

        Returns:
            Nothing.

        """
        self.font_family = font_family

    def set_font_charset(self, font_charset: int) -> None:
        """
        Set the font character set property.

        Args:
            font_charset: The font character set number.

        Returns:
            Nothing.

        """
        self.font_charset = font_charset

    def set_font_scheme(self, font_scheme: int) -> None:
        """
        Set the font scheme property.

        Args:
            font_scheme: The font scheme.

        Returns:
            Nothing.

        """
        self.font_scheme = font_scheme

    def set_font_condense(self, font_condense: int) -> None:
        """
        Set the font condense property.

        Args:
            font_condense: The font condense property.

        Returns:
            Nothing.

        """
        self.font_condense = font_condense

    def set_font_extend(self, font_extend: int) -> None:
        """
        Set the font extend property.

        Args:
            font_extend: The font extend property.

        Returns:
            Nothing.

        """
        self.font_extend = font_extend

    def set_theme(self, theme: int) -> None:
        """
        Set the theme property.

        Args:
            theme: Format theme.

        Returns:
            Nothing.

        """
        self.theme = theme

    def set_hyperlink(self, hyperlink: bool = True) -> None:
        """
        Set the properties for the hyperlink style.

        Args:
            hyperlink: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.xf_id = 1
        self.set_underline(1)
        self.set_theme(10)
        self.hyperlink = hyperlink

    def set_color_indexed(self, color_index: Literal[0, 1]) -> None:
        """
        Set the color index property. Some fundamental format properties use an
        indexed color instead of a rbg or theme color.

        Args:
            color_index: Generally 0 or 1.

        Returns:
            Nothing.

        """
        self.color_indexed = color_index

    def set_font_only(self, font_only: bool = True) -> None:
        """
        Set property to indicate that the format is used for fonts only.

        Args:
            font_only: Default is True, turns property on.

        Returns:
            Nothing.

        """
        self.font_only = font_only

    # Compatibility methods. These versions of the method names were added in an
    # initial version for compatibility testing with Excel::Writer::XLSX and
    # leaked out into production code. They are deprecated and will be removed
    # in a future after a suitable deprecation period.
    def set_font(self, font_name: str) -> None:
        """Deprecated: Use set_font_name() instead."""
        self.font_name = font_name

    def set_size(self, font_size: int) -> None:
        """Deprecated: Use set_font_size() instead."""
        self.font_size = font_size

    def set_color(self, font_color: Union[Color, str]) -> None:
        """Deprecated: Use set_font_color() instead."""
        self.font_color = Color._from_value(font_color)

    ###########################################################################
    #
    # Private API.
    #
    ###########################################################################

    def _get_align_properties(self):
        # pylint: disable=too-many-boolean-expressions
        # Return properties for an Style xf <alignment> sub-element.
        changed = 0
        align = []

        # Check if any alignment options in the format have been changed.
        if (
            self.text_h_align
            or self.text_v_align
            or self.indent
            or self.rotation
            or self.text_wrap
            or self.shrink
            or self.reading_order
        ):
            changed = 1
        else:
            return changed, align

        # Indent is only allowed for some alignment properties. If it is
        # defined for any other alignment or no alignment has been set then
        # default to left alignment.
        if (
            self.indent
            and self.text_h_align != 1
            and self.text_h_align != 3
            and self.text_h_align != 7
            and self.text_v_align != 1
            and self.text_v_align != 3
            and self.text_v_align != 5
        ):
            self.text_h_align = 1

        # Check for properties that are mutually exclusive.
        if self.text_wrap:
            self.shrink = 0
        if self.text_h_align == 4:
            self.shrink = 0
        if self.text_h_align == 5:
            self.shrink = 0
        if self.text_h_align == 7:
            self.shrink = 0
        if self.text_h_align != 7:
            self.just_distrib = 0
        if self.indent:
            self.just_distrib = 0

        continuous = "centerContinuous"

        if self.text_h_align == 1:
            align.append(("horizontal", "left"))
        if self.text_h_align == 2:
            align.append(("horizontal", "center"))
        if self.text_h_align == 3:
            align.append(("horizontal", "right"))
        if self.text_h_align == 4:
            align.append(("horizontal", "fill"))
        if self.text_h_align == 5:
            align.append(("horizontal", "justify"))
        if self.text_h_align == 6:
            align.append(("horizontal", continuous))
        if self.text_h_align == 7:
            align.append(("horizontal", "distributed"))

        if self.just_distrib:
            align.append(("justifyLastLine", 1))

        # Property 'vertical' => 'bottom' is a default. It sets applyAlignment
        # without an alignment sub-element.
        if self.text_v_align == 1:
            align.append(("vertical", "top"))
        if self.text_v_align == 2:
            align.append(("vertical", "center"))
        if self.text_v_align == 4:
            align.append(("vertical", "justify"))
        if self.text_v_align == 5:
            align.append(("vertical", "distributed"))

        if self.rotation:
            align.append(("textRotation", self.rotation))
        if self.indent:
            align.append(("indent", self.indent))

        if self.text_wrap:
            align.append(("wrapText", 1))
        if self.shrink:
            align.append(("shrinkToFit", 1))

        if self.reading_order == 1:
            align.append(("readingOrder", 1))
        if self.reading_order == 2:
            align.append(("readingOrder", 2))

        return changed, align

    def _get_protection_properties(self):
        # Return properties for an Excel XML <Protection> element.
        attributes = []

        if not self.locked:
            attributes.append(("locked", 0))
        if self.hidden:
            attributes.append(("hidden", 1))

        return attributes

    def _get_format_key(self):
        # Returns a unique hash key for a format. Used by Workbook.
        if self._format_key is None:
            self._format_key = ":".join(
                str(x)
                for x in (
                    self._get_font_key(),
                    self._get_border_key(),
                    self._get_fill_key(),
                    self._get_alignment_key(),
                    self.num_format,
                    self.locked,
                    self.checkbox,
                    self.quote_prefix,
                    self.hidden,
                )
            )

        return self._format_key

    def _get_font_key(self):
        # Returns a unique hash key for a font. Used by Workbook.
        key = ":".join(
            str(x)
            for x in (
                self.bold,
                self.font_color,
                self.font_charset,
                self.font_family,
                self.font_outline,
                self.font_script,
                self.font_shadow,
                self.font_strikeout,
                self.font_name,
                self.italic,
                self.font_size,
                self.underline,
                self.theme,
            )
        )

        return key

    def _get_border_key(self):
        # Returns a unique hash key for a border style. Used by Workbook.
        key = ":".join(
            str(x)
            for x in (
                self.bottom,
                self.bottom_color,
                self.diag_border,
                self.diag_color,
                self.diag_type,
                self.left,
                self.left_color,
                self.right,
                self.right_color,
                self.top,
                self.top_color,
            )
        )

        return key

    def _get_fill_key(self):
        # Returns a unique hash key for a fill style. Used by Workbook.
        key = ":".join(str(x) for x in (self.pattern, self.bg_color, self.fg_color))

        return key

    def _get_alignment_key(self):
        # Returns a unique hash key for alignment formats.

        key = ":".join(
            str(x)
            for x in (
                self.text_h_align,
                self.text_v_align,
                self.indent,
                self.rotation,
                self.text_wrap,
                self.shrink,
                self.reading_order,
            )
        )

        return key

    def _get_xf_index(self):
        # Returns the XF index number used by Excel to identify a format.
        if self.xf_index is not None:
            # Format already has an index number so return it.
            return self.xf_index

        # Format doesn't have an index number so assign one.
        key = self._get_format_key()

        if key in self.xf_format_indices:
            # Format matches existing format with an index.
            return self.xf_format_indices[key]

        # New format requiring an index. Note. +1 since Excel
        # has an implicit "General" format at index 0.
        index = 1 + len(self.xf_format_indices)
        self.xf_format_indices[key] = index
        self.xf_index = index
        return index

    def _get_dxf_index(self):
        # Returns the DXF index number used by Excel to identify a format.
        if self.dxf_index is not None:
            # Format already has an index number so return it.
            return self.dxf_index

        # Format doesn't have an index number so assign one.
        key = self._get_format_key()

        if key in self.dxf_format_indices:
            # Format matches existing format with an index.
            return self.dxf_format_indices[key]

        # New format requiring an index.
        index = len(self.dxf_format_indices)
        self.dxf_format_indices[key] = index
        self.dxf_index = index
        return index

    ###########################################################################
    #
    # XML methods.
    #
    ###########################################################################



##############################################################################
#
# An example of adding checkbox boolean values to a worksheet using the the
# XlsxWriter Python module.
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (c) 2013-2025, John McNamara, jmcnamara@cpan.org
#
import xlsxwriter

# Create a new Excel file object.
workbook = xlsxwriter.Workbook("checkbox.xlsx")

# Add a worksheet to the workbook.
worksheet = workbook.add_worksheet()

# Create some formats to use in the worksheet.
bold = workbook.add_format({"bold": True})
light_red = workbook.add_format({"bg_color": "#FFC7CE"})
light_green = workbook.add_format({"bg_color": "#C6EFCE"})

# Set the column width for clarity.
worksheet.set_column(0, 0, 30)

# Write some descriptions.
worksheet.write(1, 0, "Some simple checkboxes:", bold)
worksheet.write(4, 0, "Some checkboxes with cell formats:", bold)

# Insert some boolean checkboxes to the worksheet.
worksheet.insert_checkbox(1, 1, False)
worksheet.insert_checkbox(2, 1, True)

# Insert some checkboxes with cell formats.
worksheet.insert_checkbox(4, 1, False, light_red)
worksheet.insert_checkbox(5, 1, True, light_green)

# Close the workbook.
workbook.close()



###############################################################################
#
# Worksheet - A class for writing Excel Worksheets.
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (c) 2013-2025, John McNamara, jmcnamara@cpan.org
#
import datetime
import re
from typing import Dict, Optional, Tuple, Union
from warnings import warn

from xlsxwriter.color import Color

COL_NAMES: Dict[int, str] = {}

CHAR_WIDTHS = {
    " ": 3,
    "!": 5,
    '"': 6,
    "#": 7,
    "$": 7,
    "%": 11,
    "&": 10,
    "'": 3,
    "(": 5,
    ")": 5,
    "*": 7,
    "+": 7,
    ",": 4,
    "-": 5,
    ".": 4,
    "/": 6,
    "0": 7,
    "1": 7,
    "2": 7,
    "3": 7,
    "4": 7,
    "5": 7,
    "6": 7,
    "7": 7,
    "8": 7,
    "9": 7,
    ":": 4,
    ";": 4,
    "<": 7,
    "=": 7,
    ">": 7,
    "?": 7,
    "@": 13,
    "A": 9,
    "B": 8,
    "C": 8,
    "D": 9,
    "E": 7,
    "F": 7,
    "G": 9,
    "H": 9,
    "I": 4,
    "J": 5,
    "K": 8,
    "L": 6,
    "M": 12,
    "N": 10,
    "O": 10,
    "P": 8,
    "Q": 10,
    "R": 8,
    "S": 7,
    "T": 7,
    "U": 9,
    "V": 9,
    "W": 13,
    "X": 8,
    "Y": 7,
    "Z": 7,
    "[": 5,
    "\\": 6,
    "]": 5,
    "^": 7,
    "_": 7,
    "`": 4,
    "a": 7,
    "b": 8,
    "c": 6,
    "d": 8,
    "e": 8,
    "f": 5,
    "g": 7,
    "h": 8,
    "i": 4,
    "j": 4,
    "k": 7,
    "l": 4,
    "m": 12,
    "n": 8,
    "o": 8,
    "p": 8,
    "q": 8,
    "r": 5,
    "s": 6,
    "t": 5,
    "u": 8,
    "v": 7,
    "w": 11,
    "x": 7,
    "y": 7,
    "z": 6,
    "{": 5,
    "|": 7,
    "}": 5,
    "~": 7,
}

# The following is a list of Emojis used to decide if worksheet names require
# quoting since there is (currently) no native support for matching them in
# Python regular expressions. It is probably unnecessary to exclude them since
# the default quoting is safe in Excel even when unnecessary (the reverse isn't
# true). The Emoji list was generated from:
#
# https://util.unicode.org/UnicodeJsps/list-unicodeset.jsp?a=%5B%3AEmoji%3DYes%3A%5D&abb=on&esc=on&g=&i=
#
# pylint: disable-next=line-too-long
EMOJIS = "\u00a9\u00ae\u203c\u2049\u2122\u2139\u2194-\u2199\u21a9\u21aa\u231a\u231b\u2328\u23cf\u23e9-\u23f3\u23f8-\u23fa\u24c2\u25aa\u25ab\u25b6\u25c0\u25fb-\u25fe\u2600-\u2604\u260e\u2611\u2614\u2615\u2618\u261d\u2620\u2622\u2623\u2626\u262a\u262e\u262f\u2638-\u263a\u2640\u2642\u2648-\u2653\u265f\u2660\u2663\u2665\u2666\u2668\u267b\u267e\u267f\u2692-\u2697\u2699\u269b\u269c\u26a0\u26a1\u26a7\u26aa\u26ab\u26b0\u26b1\u26bd\u26be\u26c4\u26c5\u26c8\u26ce\u26cf\u26d1\u26d3\u26d4\u26e9\u26ea\u26f0-\u26f5\u26f7-\u26fa\u26fd\u2702\u2705\u2708-\u270d\u270f\u2712\u2714\u2716\u271d\u2721\u2728\u2733\u2734\u2744\u2747\u274c\u274e\u2753-\u2755\u2757\u2763\u2764\u2795-\u2797\u27a1\u27b0\u27bf\u2934\u2935\u2b05-\u2b07\u2b1b\u2b1c\u2b50\u2b55\u3030\u303d\u3297\u3299\U0001f004\U0001f0cf\U0001f170\U0001f171\U0001f17e\U0001f17f\U0001f18e\U0001f191-\U0001f19a\U0001f1e6-\U0001f1ff\U0001f201\U0001f202\U0001f21a\U0001f22f\U0001f232-\U0001f23a\U0001f250\U0001f251\U0001f300-\U0001f321\U0001f324-\U0001f393\U0001f396\U0001f397\U0001f399-\U0001f39b\U0001f39e-\U0001f3f0\U0001f3f3-\U0001f3f5\U0001f3f7-\U0001f4fd\U0001f4ff-\U0001f53d\U0001f549-\U0001f54e\U0001f550-\U0001f567\U0001f56f\U0001f570\U0001f573-\U0001f57a\U0001f587\U0001f58a-\U0001f58d\U0001f590\U0001f595\U0001f596\U0001f5a4\U0001f5a5\U0001f5a8\U0001f5b1\U0001f5b2\U0001f5bc\U0001f5c2-\U0001f5c4\U0001f5d1-\U0001f5d3\U0001f5dc-\U0001f5de\U0001f5e1\U0001f5e3\U0001f5e8\U0001f5ef\U0001f5f3\U0001f5fa-\U0001f64f\U0001f680-\U0001f6c5\U0001f6cb-\U0001f6d2\U0001f6d5-\U0001f6d7\U0001f6dc-\U0001f6e5\U0001f6e9\U0001f6eb\U0001f6ec\U0001f6f0\U0001f6f3-\U0001f6fc\U0001f7e0-\U0001f7eb\U0001f7f0\U0001f90c-\U0001f93a\U0001f93c-\U0001f945\U0001f947-\U0001f9ff\U0001fa70-\U0001fa7c\U0001fa80-\U0001fa88\U0001fa90-\U0001fabd\U0001fabf-\U0001fac5\U0001face-\U0001fadb\U0001fae0-\U0001fae8\U0001faf0-\U0001faf8"  # noqa

# Compile performance critical regular expressions.
RE_LEADING_WHITESPACE = re.compile(r"^\s")
RE_TRAILING_WHITESPACE = re.compile(r"\s$")
RE_RANGE_PARTS = re.compile(r"(\$?)([A-Z]{1,3})(\$?)(\d+)")
RE_QUOTE_RULE1 = re.compile(rf"[^\w\.{EMOJIS}]")
RE_QUOTE_RULE2 = re.compile(rf"^[\d\.{EMOJIS}]")
RE_QUOTE_RULE3 = re.compile(r"^([A-Z]{1,3}\d+)$")
RE_QUOTE_RULE4_ROW = re.compile(r"^R(\d+)")
RE_QUOTE_RULE4_COLUMN = re.compile(r"^R?C(\d+)")


def xl_rowcol_to_cell(
    row: int,
    col: int,
    row_abs: bool = False,
    col_abs: bool = False,
) -> str:
    """
    Convert a zero indexed row and column cell reference to a A1 style string.

    Args:
       row:     The cell row.    Int.
       col:     The cell column. Int.
       row_abs: Optional flag to make the row absolute.    Bool.
       col_abs: Optional flag to make the column absolute. Bool.

    Returns:
        A1 style string.

    """
    if row < 0:
        warn(f"Row number '{row}' must be >= 0")
        return ""

    if col < 0:
        warn(f"Col number '{col}' must be >= 0")
        return ""

    row += 1  # Change to 1-index.
    row_abs_str = "$" if row_abs else ""

    col_str = xl_col_to_name(col, col_abs)

    return col_str + row_abs_str + str(row)


def xl_rowcol_to_cell_fast(row: int, col: int) -> str:
    """
    Optimized version of the xl_rowcol_to_cell function. Only used internally.

    Args:
       row: The cell row.    Int.
       col: The cell column. Int.

    Returns:
        A1 style string.

    """
    if col in COL_NAMES:
        col_str = COL_NAMES[col]
    else:
        col_str = xl_col_to_name(col)
        COL_NAMES[col] = col_str

    return col_str + str(row + 1)


def xl_col_to_name(col: int, col_abs: bool = False) -> str:
    """
    Convert a zero indexed column cell reference to a string.

    Args:
       col:     The cell column. Int.
       col_abs: Optional flag to make the column absolute. Bool.

    Returns:
        Column style string.

    """
    col_num = col
    if col_num < 0:
        warn(f"Col number '{col_num}' must be >= 0")
        return ""

    col_num += 1  # Change to 1-index.
    col_str = ""
    col_abs_str = "$" if col_abs else ""

    while col_num:
        # Set remainder from 1 .. 26
        remainder = col_num % 26

        if remainder == 0:
            remainder = 26

        # Convert the remainder to a character.
        col_letter = chr(ord("A") + remainder - 1)

        # Accumulate the column letters, right to left.
        col_str = col_letter + col_str

        # Get the next order of magnitude.
        col_num = int((col_num - 1) / 26)

    return col_abs_str + col_str


def xl_cell_to_rowcol(cell_str: str) -> Tuple[int, int]:
    """
    Convert a cell reference in A1 notation to a zero indexed row and column.

    Args:
       cell_str:  A1 style string.

    Returns:
        row, col: Zero indexed cell row and column indices.

    """
    if not cell_str:
        return 0, 0

    match = RE_RANGE_PARTS.match(cell_str)
    if match is None:
        warn(f"Invalid cell reference '{cell_str}'")
        return 0, 0

    col_str = match.group(2)
    row_str = match.group(4)

    # Convert base26 column string to number.
    expn = 0
    col = 0
    for char in reversed(col_str):
        col += (ord(char) - ord("A") + 1) * (26**expn)
        expn += 1

    # Convert 1-index to zero-index
    row = int(row_str) - 1
    col -= 1

    return row, col


def xl_cell_to_rowcol_abs(cell_str: str) -> Tuple[int, int, bool, bool]:
    """
    Convert an absolute cell reference in A1 notation to a zero indexed
    row and column, with True/False values for absolute rows or columns.

    Args:
       cell_str: A1 style string.

    Returns:
        row, col, row_abs, col_abs:  Zero indexed cell row and column indices.

    """
    if not cell_str:
        return 0, 0, False, False

    match = RE_RANGE_PARTS.match(cell_str)
    if match is None:
        warn(f"Invalid cell reference '{cell_str}'")
        return 0, 0, False, False

    col_abs = bool(match.group(1))
    col_str = match.group(2)
    row_abs = bool(match.group(3))
    row_str = match.group(4)

    # Convert base26 column string to number.
    expn = 0
    col = 0
    for char in reversed(col_str):
        col += (ord(char) - ord("A") + 1) * (26**expn)
        expn += 1

    # Convert 1-index to zero-index
    row = int(row_str) - 1
    col -= 1

    return row, col, row_abs, col_abs


def xl_range(first_row: int, first_col: int, last_row: int, last_col: int) -> str:
    """
    Convert zero indexed row and col cell references to a A1:B1 range string.

    Args:
       first_row: The first cell row.    Int.
       first_col: The first cell column. Int.
       last_row:  The last cell row.     Int.
       last_col:  The last cell column.  Int.

    Returns:
        A1:B1 style range string.

    """
    range1 = xl_rowcol_to_cell(first_row, first_col)
    range2 = xl_rowcol_to_cell(last_row, last_col)

    if range1 == "" or range2 == "":
        warn("Row and column numbers must be >= 0")
        return ""

    if range1 == range2:
        return range1

    return range1 + ":" + range2


def xl_range_abs(first_row: int, first_col: int, last_row: int, last_col: int) -> str:
    """
    Convert zero indexed row and col cell references to a $A$1:$B$1 absolute
    range string.

    Args:
       first_row: The first cell row.    Int.
       first_col: The first cell column. Int.
       last_row:  The last cell row.     Int.
       last_col:  The last cell column.  Int.

    Returns:
        $A$1:$B$1 style range string.

    """
    range1 = xl_rowcol_to_cell(first_row, first_col, True, True)
    range2 = xl_rowcol_to_cell(last_row, last_col, True, True)

    if range1 == "" or range2 == "":
        warn("Row and column numbers must be >= 0")
        return ""

    if range1 == range2:
        return range1

    return range1 + ":" + range2


def xl_range_formula(
    sheetname: str, first_row: int, first_col: int, last_row: int, last_col: int
) -> str:
    """
    Convert worksheet name and zero indexed row and col cell references to
    a Sheet1!A1:B1 range formula string.

    Args:
       sheetname: The worksheet name.    String.
       first_row: The first cell row.    Int.
       first_col: The first cell column. Int.
       last_row:  The last cell row.     Int.
       last_col:  The last cell column.  Int.

    Returns:
        A1:B1 style range string.

    """
    cell_range = xl_range_abs(first_row, first_col, last_row, last_col)
    sheetname = quote_sheetname(sheetname)

    return sheetname + "!" + cell_range


def quote_sheetname(sheetname: str) -> str:
    """
    Sheetnames used in references should be quoted if they contain any spaces,
    special characters or if they look like a A1 or RC cell reference. The rules
    are shown inline below.

    Args:
       sheetname: The worksheet name. String.

    Returns:
        A quoted worksheet string.

    """
    uppercase_sheetname = sheetname.upper()
    requires_quoting = False
    col_max = 163_84
    row_max = 1048576

    # Don't quote sheetname if it is already quoted by the user.
    if not sheetname.startswith("'"):

        match_rule3 = RE_QUOTE_RULE3.match(uppercase_sheetname)
        match_rule4_row = RE_QUOTE_RULE4_ROW.match(uppercase_sheetname)
        match_rule4_column = RE_QUOTE_RULE4_COLUMN.match(uppercase_sheetname)

        # --------------------------------------------------------------------
        # Rule 1. Sheet names that contain anything other than \w and "."
        # characters must be quoted.
        # --------------------------------------------------------------------
        if RE_QUOTE_RULE1.search(sheetname):
            requires_quoting = True

        # --------------------------------------------------------------------
        # Rule 2. Sheet names that start with a digit or "." must be quoted.
        # --------------------------------------------------------------------
        elif RE_QUOTE_RULE2.search(sheetname):
            requires_quoting = True

        # --------------------------------------------------------------------
        # Rule 3. Sheet names must not be a valid A1 style cell reference.
        # Valid means that the row and column range values must also be within
        # Excel row and column limits.
        # --------------------------------------------------------------------
        elif match_rule3:
            cell = match_rule3.group(1)
            (row, col) = xl_cell_to_rowcol(cell)

            if 0 <= row < row_max and 0 <= col < col_max:
                requires_quoting = True

        # --------------------------------------------------------------------
        # Rule 4. Sheet names must not *start* with a valid RC style cell
        # reference. Other characters after the valid RC reference are ignored
        # by Excel. Valid means that the row and column range values must also
        # be within Excel row and column limits.
        #
        # Note: references without trailing characters like R12345 or C12345
        # are caught by Rule 3. Negative references like R-12345 are caught by
        # Rule 1 due to the dash.
        # --------------------------------------------------------------------

        # Rule 4a. Check for sheet names that start with R1 style references.
        elif match_rule4_row:
            row = int(match_rule4_row.group(1))

            if 0 < row <= row_max:
                requires_quoting = True

        # Rule 4b. Check for sheet names that start with C1 or RC1 style
        elif match_rule4_column:
            col = int(match_rule4_column.group(1))

            if 0 < col <= col_max:
                requires_quoting = True

        # Rule 4c. Check for some single R/C references.
        elif uppercase_sheetname in ("R", "C", "RC"):
            requires_quoting = True

    if requires_quoting:
        # Double quote any single quotes.
        sheetname = sheetname.replace("'", "''")

        # Single quote the sheet name.
        sheetname = f"'{sheetname}'"

    return sheetname


def cell_autofit_width(string: str) -> int:
    """
    Calculate the width required to auto-fit a string in a cell.

    Args:
       string: The string to calculate the cell width for. String.

    Returns:
        The string autofit width in pixels. Returns 0 if the string is empty.

    """
    if not string or len(string) == 0:
        return 0

    # Excel adds an additional 7 pixels of padding to the cell boundary.
    return xl_pixel_width(string) + 7


def xl_pixel_width(string: str) -> int:
    """
    Get the pixel width of a string based on individual character widths taken
    from Excel. UTF8 characters, and other unhandled characters, are given a
    default width of 8.

    Args:
       string: The string to calculate the width for. String.

    Returns:
        The string width in pixels. Note, Excel adds an additional 7 pixels of
        padding in the cell.

    """
    length = 0
    for char in string:
        length += CHAR_WIDTHS.get(char, 8)

    return length


def _get_sparkline_style(style_id: int) -> Dict[str, Dict[str, str]]:
    """
    Get the numbered sparkline styles.

    """
    styles = [
        {  # 0
            "low": Color.theme(4, 0),
            "high": Color.theme(4, 0),
            "last": Color.theme(4, 3),
            "first": Color.theme(4, 3),
            "series": Color.theme(4, 5),
            "markers": Color.theme(4, 5),
            "negative": Color.theme(5, 0),
        },
        {  # 1
            "low": Color.theme(4, 0),
            "high": Color.theme(4, 0),
            "last": Color.theme(4, 3),
            "first": Color.theme(4, 3),
            "series": Color.theme(4, 5),
            "markers": Color.theme(4, 5),
            "negative": Color.theme(5, 0),
        },
        {  # 2
            "low": Color.theme(5, 0),
            "high": Color.theme(5, 0),
            "last": Color.theme(5, 3),
            "first": Color.theme(5, 3),
            "series": Color.theme(5, 5),
            "markers": Color.theme(5, 5),
            "negative": Color.theme(6, 0),
        },
        {  # 3
            "low": Color.theme(6, 0),
            "high": Color.theme(6, 0),
            "last": Color.theme(6, 3),
            "first": Color.theme(6, 3),
            "series": Color.theme(6, 5),
            "markers": Color.theme(6, 5),
            "negative": Color.theme(7, 0),
        },
        {  # 4
            "low": Color.theme(7, 0),
            "high": Color.theme(7, 0),
            "last": Color.theme(7, 3),
            "first": Color.theme(7, 3),
            "series": Color.theme(7, 5),
            "markers": Color.theme(7, 5),
            "negative": Color.theme(8, 0),
        },
        {  # 5
            "low": Color.theme(8, 0),
            "high": Color.theme(8, 0),
            "last": Color.theme(8, 3),
            "first": Color.theme(8, 3),
            "series": Color.theme(8, 5),
            "markers": Color.theme(8, 5),
            "negative": Color.theme(9, 0),
        },
        {  # 6
            "low": Color.theme(9, 0),
            "high": Color.theme(9, 0),
            "last": Color.theme(9, 3),
            "first": Color.theme(9, 3),
            "series": Color.theme(9, 5),
            "markers": Color.theme(9, 5),
            "negative": Color.theme(4, 0),
        },
        {  # 7
            "low": Color.theme(5, 4),
            "high": Color.theme(5, 4),
            "last": Color.theme(5, 4),
            "first": Color.theme(5, 4),
            "series": Color.theme(4, 4),
            "markers": Color.theme(5, 4),
            "negative": Color.theme(5, 0),
        },
        {  # 8
            "low": Color.theme(6, 4),
            "high": Color.theme(6, 4),
            "last": Color.theme(6, 4),
            "first": Color.theme(6, 4),
            "series": Color.theme(5, 4),
            "markers": Color.theme(6, 4),
            "negative": Color.theme(6, 0),
        },
        {  # 9
            "low": Color.theme(7, 4),
            "high": Color.theme(7, 4),
            "last": Color.theme(7, 4),
            "first": Color.theme(7, 4),
            "series": Color.theme(6, 4),
            "markers": Color.theme(7, 4),
            "negative": Color.theme(7, 0),
        },
        {  # 10
            "low": Color.theme(8, 4),
            "high": Color.theme(8, 4),
            "last": Color.theme(8, 4),
            "first": Color.theme(8, 4),
            "series": Color.theme(7, 4),
            "markers": Color.theme(8, 4),
            "negative": Color.theme(8, 0),
        },
        {  # 11
            "low": Color.theme(9, 4),
            "high": Color.theme(9, 4),
            "last": Color.theme(9, 4),
            "first": Color.theme(9, 4),
            "series": Color.theme(8, 4),
            "markers": Color.theme(9, 4),
            "negative": Color.theme(9, 0),
        },
        {  # 12
            "low": Color.theme(4, 4),
            "high": Color.theme(4, 4),
            "last": Color.theme(4, 4),
            "first": Color.theme(4, 4),
            "series": Color.theme(9, 4),
            "markers": Color.theme(4, 4),
            "negative": Color.theme(4, 0),
        },
        {  # 13
            "low": Color.theme(4, 4),
            "high": Color.theme(4, 4),
            "last": Color.theme(4, 4),
            "first": Color.theme(4, 4),
            "series": Color.theme(4, 0),
            "markers": Color.theme(4, 4),
            "negative": Color.theme(5, 0),
        },
        {  # 14
            "low": Color.theme(5, 4),
            "high": Color.theme(5, 4),
            "last": Color.theme(5, 4),
            "first": Color.theme(5, 4),
            "series": Color.theme(5, 0),
            "markers": Color.theme(5, 4),
            "negative": Color.theme(6, 0),
        },
        {  # 15
            "low": Color.theme(6, 4),
            "high": Color.theme(6, 4),
            "last": Color.theme(6, 4),
            "first": Color.theme(6, 4),
            "series": Color.theme(6, 0),
            "markers": Color.theme(6, 4),
            "negative": Color.theme(7, 0),
        },
        {  # 16
            "low": Color.theme(7, 4),
            "high": Color.theme(7, 4),
            "last": Color.theme(7, 4),
            "first": Color.theme(7, 4),
            "series": Color.theme(7, 0),
            "markers": Color.theme(7, 4),
            "negative": Color.theme(8, 0),
        },
        {  # 17
            "low": Color.theme(8, 4),
            "high": Color.theme(8, 4),
            "last": Color.theme(8, 4),
            "first": Color.theme(8, 4),
            "series": Color.theme(8, 0),
            "markers": Color.theme(8, 4),
            "negative": Color.theme(9, 0),
        },
        {  # 18
            "low": Color.theme(9, 4),
            "high": Color.theme(9, 4),
            "last": Color.theme(9, 4),
            "first": Color.theme(9, 4),
            "series": Color.theme(9, 0),
            "markers": Color.theme(9, 4),
            "negative": Color.theme(4, 0),
        },
        {  # 19
            "low": Color.theme(4, 5),
            "high": Color.theme(4, 5),
            "last": Color.theme(4, 4),
            "first": Color.theme(4, 4),
            "series": Color.theme(4, 3),
            "markers": Color.theme(4, 1),
            "negative": Color.theme(0, 5),
        },
        {  # 20
            "low": Color.theme(5, 5),
            "high": Color.theme(5, 5),
            "last": Color.theme(5, 4),
            "first": Color.theme(5, 4),
            "series": Color.theme(5, 3),
            "markers": Color.theme(5, 1),
            "negative": Color.theme(0, 5),
        },
        {  # 21
            "low": Color.theme(6, 5),
            "high": Color.theme(6, 5),
            "last": Color.theme(6, 4),
            "first": Color.theme(6, 4),
            "series": Color.theme(6, 3),
            "markers": Color.theme(6, 1),
            "negative": Color.theme(0, 5),
        },
        {  # 22
            "low": Color.theme(7, 5),
            "high": Color.theme(7, 5),
            "last": Color.theme(7, 4),
            "first": Color.theme(7, 4),
            "series": Color.theme(7, 3),
            "markers": Color.theme(7, 1),
            "negative": Color.theme(0, 5),
        },
        {  # 23
            "low": Color.theme(8, 5),
            "high": Color.theme(8, 5),
            "last": Color.theme(8, 4),
            "first": Color.theme(8, 4),
            "series": Color.theme(8, 3),
            "markers": Color.theme(8, 1),
            "negative": Color.theme(0, 5),
        },
        {  # 24
            "low": Color.theme(9, 5),
            "high": Color.theme(9, 5),
            "last": Color.theme(9, 4),
            "first": Color.theme(9, 4),
            "series": Color.theme(9, 3),
            "markers": Color.theme(9, 1),
            "negative": Color.theme(0, 5),
        },
        {  # 25
            "low": Color.theme(1, 3),
            "high": Color.theme(1, 3),
            "last": Color.theme(1, 3),
            "first": Color.theme(1, 3),
            "series": Color.theme(1, 1),
            "markers": Color.theme(1, 3),
            "negative": Color.theme(1, 3),
        },
        {  # 26
            "low": Color.theme(0, 3),
            "high": Color.theme(0, 3),
            "last": Color.theme(0, 3),
            "first": Color.theme(0, 3),
            "series": Color.theme(1, 2),
            "markers": Color.theme(0, 3),
            "negative": Color.theme(0, 3),
        },
        {  # 27
            "low": Color("#D00000"),
            "high": Color("#D00000"),
            "last": Color("#D00000"),
            "first": Color("#D00000"),
            "series": Color("#323232"),
            "markers": Color("#D00000"),
            "negative": Color("#D00000"),
        },
        {  # 28
            "low": Color("#0070C0"),
            "high": Color("#0070C0"),
            "last": Color("#0070C0"),
            "first": Color("#0070C0"),
            "series": Color("#000000"),
            "markers": Color("#0070C0"),
            "negative": Color("#0070C0"),
        },
        {  # 29
            "low": Color("#D00000"),
            "high": Color("#D00000"),
            "last": Color("#D00000"),
            "first": Color("#D00000"),
            "series": Color("#376092"),
            "markers": Color("#D00000"),
            "negative": Color("#D00000"),
        },
        {  # 30
            "low": Color("#000000"),
            "high": Color("#000000"),
            "last": Color("#000000"),
            "first": Color("#000000"),
            "series": Color("#0070C0"),
            "markers": Color("#000000"),
            "negative": Color("#000000"),
        },
        {  # 31
            "low": Color("#FF5055"),
            "high": Color("#56BE79"),
            "last": Color("#359CEB"),
            "first": Color("#5687C2"),
            "series": Color("#5F5F5F"),
            "markers": Color("#D70077"),
            "negative": Color("#FFB620"),
        },
        {  # 32
            "low": Color("#FF5055"),
            "high": Color("#56BE79"),
            "last": Color("#359CEB"),
            "first": Color("#777777"),
            "series": Color("#5687C2"),
            "markers": Color("#D70077"),
            "negative": Color("#FFB620"),
        },
        {  # 33
            "low": Color("#FF5367"),
            "high": Color("#60D276"),
            "last": Color("#FFEB9C"),
            "first": Color("#FFDC47"),
            "series": Color("#C6EFCE"),
            "markers": Color("#8CADD6"),
            "negative": Color("#FFC7CE"),
        },
        {  # 34
            "low": Color("#FF0000"),
            "high": Color("#00B050"),
            "last": Color("#FFC000"),
            "first": Color("#FFC000"),
            "series": Color("#00B050"),
            "markers": Color("#0070C0"),
            "negative": Color("#FF0000"),
        },
        {  # 35
            "low": Color.theme(7, 0),
            "high": Color.theme(6, 0),
            "last": Color.theme(5, 0),
            "first": Color.theme(4, 0),
            "series": Color.theme(3, 0),
            "markers": Color.theme(8, 0),
            "negative": Color.theme(9, 0),
        },
        {  # 36
            "low": Color.theme(7, 0),
            "high": Color.theme(6, 0),
            "last": Color.theme(5, 0),
            "first": Color.theme(4, 0),
            "series": Color.theme(1, 0),
            "markers": Color.theme(8, 0),
            "negative": Color.theme(9, 0),
        },
    ]

    return styles[style_id]


def _supported_datetime(
    dt: Union[datetime.datetime, datetime.time, datetime.date],
) -> bool:
    # Determine is an argument is a supported datetime object.
    return isinstance(
        dt, (datetime.datetime, datetime.date, datetime.time, datetime.timedelta)
    )


def _remove_datetime_timezone(
    dt_obj: datetime.datetime, remove_timezone: bool
) -> datetime.datetime:
    # Excel doesn't support timezones in datetimes/times so we remove the
    # tzinfo from the object if the user has specified that option in the
    # constructor.
    if remove_timezone:
        dt_obj = dt_obj.replace(tzinfo=None)
    else:
        if dt_obj.tzinfo:
            raise TypeError(
                "Excel doesn't support timezones in datetimes. "
                "Set the tzinfo in the datetime/time object to None or "
                "use the 'remove_timezone' Workbook() option"
            )

    return dt_obj


def _datetime_to_excel_datetime(
    dt_obj: Union[datetime.time, datetime.datetime, datetime.timedelta, datetime.date],
    date_1904: bool,
    remove_timezone: bool,
) -> float:
    # Convert a datetime object to an Excel serial date and time. The integer
    # part of the number stores the number of days since the epoch and the
    # fractional part stores the percentage of the day.
    date_type = dt_obj
    is_timedelta = False

    if date_1904:
        # Excel for Mac date epoch.
        epoch = datetime.datetime(1904, 1, 1)
    else:
        # Default Excel epoch.
        epoch = datetime.datetime(1899, 12, 31)

    # We handle datetime .datetime, .date and .time objects but convert
    # them to datetime.datetime objects and process them in the same way.
    if isinstance(dt_obj, datetime.datetime):
        dt_obj = _remove_datetime_timezone(dt_obj, remove_timezone)
        delta = dt_obj - epoch
    elif isinstance(dt_obj, datetime.date):
        dt_obj = datetime.datetime.fromordinal(dt_obj.toordinal())
        delta = dt_obj - epoch
    elif isinstance(dt_obj, datetime.time):
        dt_obj = datetime.datetime.combine(epoch, dt_obj)
        dt_obj = _remove_datetime_timezone(dt_obj, remove_timezone)
        delta = dt_obj - epoch
    elif isinstance(dt_obj, datetime.timedelta):
        is_timedelta = True
        delta = dt_obj
    else:
        raise TypeError("Unknown or unsupported datetime type")

    # Convert a Python datetime.datetime value to an Excel date number.
    excel_time = delta.days + (
        float(delta.seconds) + float(delta.microseconds) / 1e6
    ) / (60 * 60 * 24)

    # The following is a workaround for the fact that in Excel a time only
    # value is represented as 1899-12-31+time whereas in datetime.datetime()
    # it is 1900-1-1+time so we need to subtract the 1 day difference.
    if (
        isinstance(date_type, datetime.datetime)
        and not isinstance(dt_obj, datetime.timedelta)
        and dt_obj.isocalendar()
        == (
            1900,
            1,
            1,
        )
    ):
        excel_time -= 1

    # Account for Excel erroneously treating 1900 as a leap year.
    if not date_1904 and not is_timedelta and excel_time > 59:
        excel_time += 1

    return excel_time


def _preserve_whitespace(string: str) -> Optional[re.Match]:
    # Check if a string has leading or trailing whitespace that requires a
    # "preserve" attribute.
    return RE_LEADING_WHITESPACE.search(string) or RE_TRAILING_WHITESPACE.search(string)



##############################################################################
#
# A demonstration of the RGB and Theme colors palettes available in the
# XlsxWriter Python module.
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (c) 2013-2025, John McNamara, jmcnamara@cpan.org
#
import xlsxwriter
from xlsxwriter.color import Color

# Create a new Excel file object.
workbook = xlsxwriter.Workbook("colors.xlsx")

# Add a worksheet for the RGB colors.
worksheet = workbook.add_worksheet("RGB Colors")

# Write some predefined colors to cells.
named_colors = [
    "Black",
    "Blue",
    "Brown",
    "Cyan",
    "Gray",
    "Green",
    "Lime",
    "Magenta",
    "Navy",
    "Orange",
    "Pink",
    "Purple",
    "Red",
    "Silver",
    "White",
    "Yellow",
]

# Write the named colors.
for row, name in enumerate(named_colors):
    color_format = workbook.add_format({"bg_color": Color(name)})
    worksheet.write_string(row, 0, name)
    worksheet.write_blank(row, 1, None, color_format)

# Write some user-defined RGB colors to cells.
user_defined_colors = [
    "#FF7F50",
    "#DCDCDC",
    "#6495ED",
    "#DAA520",
]

for row, name in enumerate(user_defined_colors, start=len(named_colors)):
    color_format = workbook.add_format({"bg_color": Color(name)})
    worksheet.write_string(row, 0, name)
    worksheet.write_blank(row, 1, None, color_format)

# Add a worksheet for the Theme colors.
worksheet = workbook.add_worksheet("Theme Colors")

# Add alternative colors for the cell text.
black_text = Color("#000000")
white_text = Color("#FFFFFF")

# Create a cell with each of the theme colors.
for row in range(6):
    for col in range(10):
        # Use the theme color for the background.
        theme_color = Color.theme(col, row)

        # Use white font color for better contrast on dark backgrounds.
        if col != 0:
            font_color = white_text
        else:
            font_color = black_text

        color_format = workbook.add_format(
            {
                "align": "center",
                "bg_color": theme_color,
                "font_color": font_color,
            }
        )
        worksheet.write_string(row, col, f"({col}, {row})", color_format)

# Save the file to disk.
workbook.close()



##############################################################################
#
# The following example demonstrates manually auto-fitting the the width of a
# column in Excel based on the maximum string width. The worksheet ``autofit()``
# method will do this automatically but occasionally you may need to control the
# maximum and minimum column widths yourself.
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (c) 2013-2025, John McNamara, jmcnamara@cpan.org
#
from functools import reduce

import xlsxwriter
from xlsxwriter.utility import cell_autofit_width

workbook = xlsxwriter.Workbook("autofit.xlsx")
worksheet = workbook.add_worksheet()

# Some string data to write.
cities = ["Addis Ababa", "Buenos Aires", "Cairo", "Dhaka"]

# Write the strings:
worksheet.write_column(0, 0, cities)

# Find the maximum column width in pixels.
max_width = reduce(max, map(cell_autofit_width, cities))

# Set the column width as if it was auto-fitted.
worksheet.set_column_pixels(0, 0, max_width)

workbook.close()



##############################################################################
#
# An example of adding a Sensitivity Label to an XlsxWriter file.
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (c) 2013-2025, John McNamara, jmcnamara@cpan.org
#
import xlsxwriter

workbook = xlsxwriter.Workbook("sensitivity_label.xlsx")
worksheet = workbook.add_worksheet()

# Metadata extracted from a company specific file.
company_guid = "2096f6a2-d2f7-48be-b329-b73aaa526e5d"
site_id = "cb46c030-1825-4e81-a295-151c039dbf02"
action_id = "88124cf5-1340-457d-90e1-0000a9427c99"

# Add the document properties. Note that these should all be in text format.
workbook.set_custom_property(f"MSIP_Label_{company_guid}_Enabled", "true", "text")
workbook.set_custom_property(
    f"MSIP_Label_{company_guid}_SetDate", "2024-01-01T12:00:00Z", "text"
)
workbook.set_custom_property(f"MSIP_Label_{company_guid}_Method", "Privileged", "text")
workbook.set_custom_property(f"MSIP_Label_{company_guid}_Name", "Confidential", "text")
workbook.set_custom_property(f"MSIP_Label_{company_guid}_SiteId", site_id, "text")
workbook.set_custom_property(f"MSIP_Label_{company_guid}_ActionId", action_id, "text")
workbook.set_custom_property(f"MSIP_Label_{company_guid}_ContentBits", "2", "text")

workbook.close()



import sys
from warnings import warn

from setuptools import setup

if sys.version_info < (3, 8):
    warn("The minimum Python version supported by XlsxWriter is 3.8")
    sys.exit()

setup(
    name="xlsxwriter",
    version="3.2.8",
    author="John McNamara",
    author_email="jmcnamara@cpan.org",
    url="https://github.com/jmcnamara/XlsxWriter",
    packages=["xlsxwriter"],
    scripts=["examples/vba_extract.py"],
    license="BSD-2-Clause",
    description="A Python module for creating Excel XLSX files.",
    long_description=open("README.rst", encoding="utf-8").read(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
    python_requires=">=3.8",
)
